import logging
from collections import defaultdict
from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Sequence, Set, Tuple

import h5py
import numpy as np
from tqdm import tqdm

import fast_forward
from fast_forward.encoder import Encoder
from fast_forward.index import IDSequence, Index, Mode
from fast_forward.index.memory import InMemoryIndex
from fast_forward.quantizer import Quantizer

LOGGER = logging.getLogger(__name__)


class OnDiskIndex(Index):
    """Fast-Forward index that is read on-demand from disk.

    Uses HDF5 via h5py under the hood. The buffer (`ds_buffer_size`) works around a [h5py limitation](https://docs.h5py.org/en/latest/high/dataset.html#fancy-indexing).
    """

    def __init__(
        self,
        index_file: Path,
        query_encoder: Encoder = None,
        quantizer: Quantizer = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
        init_size: int = 2**14,
        resize_min_val: int = 2**10,
        hdf5_chunk_size: int = None,
        max_id_length: int = 8,
        overwrite: bool = False,
        ds_buffer_size: int = 2**10,
    ) -> None:
        """Create an index.

        Args:
            index_file (Path): Index file to create (or overwrite).
            query_encoder (Encoder, optional): Query encoder. Defaults to None.
            quantizer (Quantizer, optional): The quantizer to use. Defaults to None.
            mode (Mode, optional): Ranking mode. Defaults to Mode.MAXP.
            encoder_batch_size (int, optional): Batch size for query encoder. Defaults to 32.
            init_size (int, optional): Initial size to allocate (number of vectors). Defaults to 2**14.
            resize_min_val (int, optional): Minimum number of vectors to increase index size by. Defaults to 2**10.
            hdf5_chunk_size (int, optional): Override chunk size used by HDF5. Defaults to None.
            max_id_length (int, optional): Maximum length of document and passage IDs (number of characters). Defaults to 8.
            overwrite (bool, optional): Overwrite index file if it exists. Defaults to False.
            ds_buffer_size (int, optional): Maximum number of vectors to retrieve from the HDF5 dataset at once. Defaults to 2**10.

        Raises:
            ValueError: When the file exists and `overwrite=False`.
        """
        if index_file.exists() and not overwrite:
            raise ValueError(f"File {index_file} exists.")

        self._index_file = index_file.absolute()
        self._doc_id_to_idx = defaultdict(list)
        self._psg_id_to_idx = {}

        self._init_size = init_size
        self._resize_min_val = resize_min_val
        self._hdf5_chunk_size = hdf5_chunk_size
        self._max_id_length = max_id_length
        self._ds_buffer_size = ds_buffer_size

        LOGGER.debug("creating file %s", self._index_file)
        with h5py.File(self._index_file, "w") as fp:
            fp.attrs["num_vectors"] = 0
            fp.attrs["ff_version"] = fast_forward.__version__

        super().__init__(
            query_encoder=query_encoder,
            quantizer=quantizer,
            mode=mode,
            encoder_batch_size=encoder_batch_size,
        )

    @Index.quantizer.setter
    def quantizer(self, quantizer: Quantizer) -> None:
        # call the setter of the super class
        Index.quantizer.fset(self, quantizer)

        # serialize the quantizer and store it on disk
        with h5py.File(self._index_file, "a") as fp:
            if "quantizer" in fp:
                del fp["quantizer"]

            meta, attributes, data = self.quantizer.serialize()
            fp.create_group("quantizer/meta").attrs.update(meta)
            fp.create_group("quantizer/attributes").attrs.update(attributes)
            data_group = fp.create_group("quantizer/data")
            for k, v in data.items():
                data_group.create_dataset(k, data=v)

    def _create_ds(self, fp: h5py.File, dim: int, dtype: np.dtype) -> None:
        """Create the HDF5 datasets for vectors and IDs.

        Args:
            fp (h5py.File): Index file (write permissions).
            dim (int): Dimension of the vectors.
            dtype (np.dtype): Type of the vectors.
        """
        fp.create_dataset(
            "vectors",
            (self._init_size, dim),
            dtype,
            maxshape=(None, dim),
            chunks=(
                True if self._hdf5_chunk_size is None else (self._hdf5_chunk_size, dim)
            ),
        )
        fp.create_dataset(
            "doc_ids",
            (self._init_size,),
            f"S{self._max_id_length}",
            maxshape=(None,),
            chunks=(
                True if self._hdf5_chunk_size is None else (self._hdf5_chunk_size,)
            ),
        )
        fp.create_dataset(
            "psg_ids",
            (self._init_size,),
            f"S{self._max_id_length}",
            maxshape=(None,),
            chunks=(
                True if self._hdf5_chunk_size is None else (self._hdf5_chunk_size,)
            ),
        )

    def __len__(self) -> int:
        with h5py.File(self._index_file, "r") as fp:
            return fp.attrs["num_vectors"]

    def _get_internal_dim(self) -> Optional[int]:
        with h5py.File(self._index_file, "r") as fp:
            if "vectors" in fp:
                return fp["vectors"].shape[1]
        return None

    def to_memory(self, buffer_size=None) -> InMemoryIndex:
        """Load the index entirely into memory.

        Args:
            buffer_size (int, optional): Use batches instead of adding all vectors at once. Defaults to None.

        Returns:
            InMemoryIndex: The loaded index.
        """
        index = InMemoryIndex(
            query_encoder=self._query_encoder,
            quantizer=self._quantizer,
            mode=self.mode,
            encoder_batch_size=self._encoder_batch_size,
            init_size=len(self),
        )
        with h5py.File(self._index_file, "r") as fp:
            buffer_size = buffer_size or fp.attrs["num_vectors"]
            for i_low in range(0, fp.attrs["num_vectors"], buffer_size):
                i_up = min(i_low + buffer_size, fp.attrs["num_vectors"])

                # IDs that don't exist will be returned as empty strings here
                doc_ids = fp["doc_ids"].asstr()[i_low:i_up]
                doc_ids[doc_ids == ""] = None
                psg_ids = fp["psg_ids"].asstr()[i_low:i_up]
                psg_ids[psg_ids == ""] = None
                index._add(fp["vectors"][i_low:i_up], doc_ids=doc_ids, psg_ids=psg_ids)
        return index

    def _add(
        self,
        vectors: np.ndarray,
        doc_ids: Sequence[Optional[str]],
        psg_ids: Sequence[Optional[str]],
    ) -> None:
        with h5py.File(self._index_file, "a") as fp:
            # if this is the first call to _add, no datasets exist
            if "vectors" not in fp:
                self._create_ds(fp, vectors.shape[-1], vectors.dtype)

            # check all IDs first before adding anything
            doc_id_size = fp["doc_ids"].dtype.itemsize
            for doc_id in doc_ids:
                if doc_id is not None and len(doc_id) > doc_id_size:
                    raise RuntimeError(
                        f"Document ID {doc_id} is longer than the maximum ({doc_id_size} characters)."
                    )
            psg_id_size = fp["psg_ids"].dtype.itemsize
            for psg_id in psg_ids:
                if psg_id is not None and len(psg_id) > psg_id_size:
                    raise RuntimeError(
                        f"Passage ID {psg_id} is longer than the maximum ({psg_id_size} characters)."
                    )

            num_new_vecs = vectors.shape[0]
            capacity = fp["vectors"].shape[0]

            # check if we have enough space, resize if necessary
            cur_num_vectors = fp.attrs["num_vectors"]
            space_left = capacity - cur_num_vectors
            if num_new_vecs > space_left:
                new_size = max(
                    capacity + num_new_vecs - space_left, self._resize_min_val
                )
                LOGGER.debug("resizing index from %s to %s", capacity, new_size)
                fp["vectors"].resize(new_size, axis=0)
                fp["doc_ids"].resize(new_size, axis=0)
                fp["psg_ids"].resize(new_size, axis=0)

            # add new document IDs to index and in-memory mappings
            doc_id_idxs, non_null_doc_ids = [], []
            for i, doc_id in enumerate(doc_ids):
                if doc_id is not None:
                    self._doc_id_to_idx[doc_id].append(cur_num_vectors + i)
                    doc_id_idxs.append(cur_num_vectors + i)
                    non_null_doc_ids.append(doc_id)
            fp["doc_ids"][doc_id_idxs] = non_null_doc_ids

            # add new passage IDs to index and in-memory mappings
            psg_id_idxs, non_null_psg_ids = [], []
            for i, psg_id in enumerate(psg_ids):
                if psg_id is not None:
                    self._psg_id_to_idx[psg_id] = cur_num_vectors + i
                    psg_id_idxs.append(cur_num_vectors + i)
                    non_null_psg_ids.append(psg_id)
            fp["psg_ids"][psg_id_idxs] = non_null_psg_ids

            # add new vectors
            fp["vectors"][cur_num_vectors : cur_num_vectors + num_new_vecs] = vectors
            fp.attrs["num_vectors"] += num_new_vecs

    @property
    def doc_ids(self) -> Set[str]:
        return set(self._doc_id_to_idx.keys())

    @property
    def psg_ids(self) -> Set[str]:
        return set(self._psg_id_to_idx.keys())

    def _get_vectors(self, ids: Iterable[str]) -> Tuple[np.ndarray, List[List[int]]]:
        idx_pairs = []
        with h5py.File(self._index_file, "r") as fp:
            for id in ids:
                if self.mode in (Mode.MAXP, Mode.AVEP) and id in self._doc_id_to_idx:
                    idxs = self._doc_id_to_idx[id]
                elif self.mode == Mode.FIRSTP and id in self._doc_id_to_idx:
                    idxs = [self._doc_id_to_idx[id][0]]
                elif self.mode == Mode.PASSAGE and id in self._psg_id_to_idx:
                    idxs = [self._psg_id_to_idx[id]]
                else:
                    LOGGER.warning("no vectors for %s", id)
                    idxs = []

                for idx in idxs:
                    idx_pairs.append((id, idx))

            # h5py requires accessing the dataset with sorted indices
            idx_pairs.sort(key=lambda x: x[1])
            id_to_idxs = defaultdict(list)
            vec_idxs = []
            for id_idx, (id, vec_idx) in enumerate(idx_pairs):
                vec_idxs.append(vec_idx)
                id_to_idxs[id].append(id_idx)

            # reading all vectors at once slows h5py down significantly, so we read them in chunks and concatenate
            vectors = np.concatenate(
                [
                    fp["vectors"][vec_idxs[i : i + self._ds_buffer_size]]
                    for i in range(0, len(vec_idxs), self._ds_buffer_size)
                ]
            )
            return vectors, [id_to_idxs[id] for id in ids]

    def _batch_iter(
        self, batch_size: int
    ) -> Iterator[Tuple[np.ndarray, IDSequence, IDSequence]]:
        with h5py.File(self._index_file, "r") as fp:
            num_vectors = fp.attrs["num_vectors"]
            for i in range(0, num_vectors, batch_size):
                j = min(i + batch_size, num_vectors)
                doc_ids = fp["doc_ids"].asstr()[i:j]
                doc_ids[doc_ids == ""] = None
                psg_ids = fp["psg_ids"].asstr()[i:j]
                psg_ids[psg_ids == ""] = None
                yield (
                    fp["vectors"][i:j],
                    doc_ids.tolist(),
                    psg_ids.tolist(),
                )

    @classmethod
    def load(
        cls,
        index_file: Path,
        query_encoder: Encoder = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
        resize_min_val: int = 2**10,
        ds_buffer_size: int = 2**10,
    ) -> "OnDiskIndex":
        """Open an existing index on disk.

        Args:
            index_file (Path): Index file to open.
            query_encoder (Encoder, optional): Query encoder. Defaults to None.
            mode (Mode, optional): Ranking mode. Defaults to Mode.MAXP.
            encoder_batch_size (int, optional): Batch size for query encoder. Defaults to 32.
            resize_min_val (int, optional): Minimum value to increase index size by. Defaults to 2**10.
            ds_buffer_size (int, optional): Maximum number of vectors to retrieve from the HDF5 dataset at once. Defaults to 2**10.

        Returns:
            OnDiskIndex: The index.
        """
        LOGGER.debug("reading file %s", index_file)

        index = cls.__new__(cls)
        super(OnDiskIndex, index).__init__(
            query_encoder=query_encoder,
            quantizer=None,
            mode=mode,
            encoder_batch_size=encoder_batch_size,
        )
        index._index_file = index_file.absolute()
        index._resize_min_val = resize_min_val
        index._ds_buffer_size = ds_buffer_size

        # deserialize quantizer if any
        with h5py.File(index_file, "r") as fp:
            if "quantizer" in fp:
                index._quantizer = Quantizer.deserialize(
                    dict(fp["quantizer/meta"].attrs),
                    dict(fp["quantizer/attributes"].attrs),
                    {k: v[:] for k, v in fp["quantizer/data"].items()},
                )

        # read ID mappings
        with h5py.File(index_file, "r") as fp:
            index._doc_id_to_idx = defaultdict(list)
            index._psg_id_to_idx = {}

            num_vectors = fp.attrs["num_vectors"]
            if num_vectors == 0:
                return index

            doc_id_iter = fp["doc_ids"].asstr()[:num_vectors]
            psg_id_iter = fp["psg_ids"].asstr()[:num_vectors]
            for i, (doc_id, psg_id) in tqdm(
                enumerate(zip(doc_id_iter, psg_id_iter)),
                total=num_vectors,
            ):
                if len(doc_id) > 0:
                    index._doc_id_to_idx[doc_id].append(i)
                if len(psg_id) > 0:
                    index._psg_id_to_idx[psg_id] = i
        return index



import os
import tempfile
import unittest
from pathlib import Path

import numpy as np
import pandas as pd

from fast_forward import Ranking

RUN = {
    "q1": {"d0": 1, "d1": 2, "d2": 300},
    "q2": {"d0": 4, "d1": 5, "d2": 600, "d3": 7},
}
DUMMY_QUERIES = {"q1": "query 1", "q2": "query 2"}


class TestRanking(unittest.TestCase):
    def test_ranking(self):
        r = Ranking.from_run(RUN)
        self.assertEqual({"q1", "q2"}, r.q_ids)
        self.assertEqual(len(r), 2)
        self.assertIn("q1", r)
        self.assertIn("q2", r)
        self.assertNotIn("q3", r)

    def test_attach_queries(self):
        r = Ranking.from_run(RUN)
        r_with_queries = r.attach_queries(DUMMY_QUERIES)
        self.assertFalse(r.has_queries)
        self.assertTrue(r_with_queries.has_queries)
        self.assertEqual(
            pd.unique(r_with_queries._df.loc[r._df["q_id"].eq("q1"), "query"]).tolist(),
            ["query 1"],
        )
        self.assertEqual(
            pd.unique(r_with_queries._df.loc[r._df["q_id"].eq("q2"), "query"]).tolist(),
            ["query 2"],
        )

        more_queries = {"qx": "other query"}
        with self.assertRaises(ValueError):
            Ranking.from_run(RUN, queries=more_queries)

        more_queries.update(DUMMY_QUERIES)
        r2_with_queries = Ranking.from_run(RUN, queries=more_queries)
        self.assertEqual(r2_with_queries, r_with_queries)

    def test_eq(self):
        r1 = Ranking.from_run({"q1": {"d1": 1, "d2": 2}})
        r2 = Ranking.from_run({"q1": {"d2": 2, "d1": 1}})
        r3 = Ranking.from_run({"q1": {"d1": 2, "d2": 3}})
        r4 = Ranking.from_run({"q1": {"d1": 1, "d2": 2}, "q2": {}})
        self.assertEqual(r1, r2)
        self.assertNotEqual(r1, r3)
        self.assertEqual(r1, r4)

        self.assertNotEqual(r1, "string")
        self.assertNotEqual(r1, 0)

    def test_cut(self):
        self.assertEqual(
            Ranking.from_run(RUN).cut(2),
            Ranking.from_run({"q1": {"d2": 300, "d1": 2}, "q2": {"d2": 600, "d3": 7}}),
        )

    def test_save_load(self):
        r = Ranking.from_run(RUN, name="Dummy")
        fd, f = tempfile.mkstemp()
        f = Path(f)
        r.save(f)
        r_from_file = Ranking.from_file(f)
        self.assertEqual(r, r_from_file)
        self.assertEqual(r.name, r_from_file.name)
        os.close(fd)
        os.remove(f)

    def test_interpolate(self):
        r = Ranking.from_run(RUN)
        df = r._df.copy()
        df["score"] = list(map(np.float32, range(len(r._df))))
        r2 = Ranking(df)
        r_int = r.interpolate(r2, 0.5)
        self.assertNotEqual(r, r_int)
        self.assertEqual(r_int["q1"], {"d2": 152.0, "d1": 3.5, "d0": 3.5})
        self.assertEqual(r_int["q2"], {"d2": 300.0, "d3": 4.0, "d1": 3.5, "d0": 3.5})


if __name__ == "__main__":
    unittest.main()



"""
.. include:: docs/encoder.md
"""

import abc
from pathlib import Path
from typing import Callable, Sequence, Union

import numpy as np
import torch
from transformers import AutoModel, AutoTokenizer


class Encoder(abc.ABC):
    """Base class for encoders."""

    @abc.abstractmethod
    def __call__(self, texts: Sequence[str]) -> np.ndarray:
        """Encode a list of texts.

        Args:
            texts (Sequence[str]): The texts to encode.

        Returns:
            np.ndarray: The resulting vector representations.
        """
        pass


class TransformerEncoder(Encoder):
    """Uses a pre-trained transformer model for encoding. Returns the pooler output."""

    def __init__(
        self, model: Union[str, Path], device: str = "cpu", **tokenizer_args
    ) -> None:
        """Create a transformer encoder.

        Args:
            model (Union[str, Path]): Pre-trained transformer model (name or path).
            device (str, optional): PyTorch device. Defaults to "cpu".
            **tokenizer_args: Additional tokenizer arguments.
        """
        super().__init__()
        self.model = AutoModel.from_pretrained(model)
        self.model.to(device)
        self.model.eval()
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self.device = device
        self.tokenizer_args = tokenizer_args

    def __call__(self, texts: Sequence[str]) -> np.ndarray:
        inputs = self.tokenizer(texts, return_tensors="pt", **self.tokenizer_args)
        inputs.to(self.device)
        with torch.no_grad():
            return self.model(**inputs).pooler_output.detach().cpu().numpy()


class LambdaEncoder(Encoder):
    """Encoder adapter class for arbitrary encoding functions."""

    def __init__(self, f: Callable[[str], np.ndarray]) -> None:
        """Create a lambda encoder.

        Args:
            f (Callable[[str], np.ndarray]): Function to encode a single piece of text.
        """
        super().__init__()
        self._f = f

    def __call__(self, texts: Sequence[str]) -> np.ndarray:
        return np.array(list(map(self._f, texts)))


class TCTColBERTQueryEncoder(TransformerEncoder):
    """Query encoder for pre-trained TCT-ColBERT models.

    Adapted from Pyserini:
    https://github.com/castorini/pyserini/blob/310c828211bb3b9528cfd59695184c80825684a2/pyserini/encode/_tct_colbert.py#L72
    """

    def __call__(self, queries: Sequence[str]) -> np.ndarray:
        max_length = 36
        inputs = self.tokenizer(
            ["[CLS] [Q] " + q + "[MASK]" * max_length for q in queries],
            max_length=max_length,
            truncation=True,
            add_special_tokens=False,
            return_tensors="pt",
            **self.tokenizer_args,
        )
        inputs.to(self.device)
        with torch.no_grad():
            embeddings = self.model(**inputs).last_hidden_state.detach().cpu().numpy()
            return np.average(embeddings[:, 4:, :], axis=-2)


class TCTColBERTDocumentEncoder(TransformerEncoder):
    """Document encoder for pre-trained TCT-ColBERT models.

    Adapted from Pyserini:
    https://github.com/castorini/pyserini/blob/310c828211bb3b9528cfd59695184c80825684a2/pyserini/encode/_tct_colbert.py#L27
    """

    def __call__(self, texts: Sequence[str]) -> np.ndarray:
        max_length = 512
        inputs = self.tokenizer(
            ["[CLS] [D] " + text for text in texts],
            max_length=max_length,
            padding=True,
            truncation=True,
            add_special_tokens=False,
            return_tensors="pt",
            **self.tokenizer_args,
        )
        inputs.to(self.device)
        with torch.no_grad():
            outputs = self.model(**inputs)
            token_embeddings = outputs["last_hidden_state"][:, 4:, :]
            input_mask_expanded = (
                inputs["attention_mask"][:, 4:]
                .unsqueeze(-1)
                .expand(token_embeddings.size())
                .float()
            )
            sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, 1)
            sum_mask = torch.clamp(input_mask_expanded.sum(1), min=1e-9)
            embeddings = sum_embeddings / sum_mask
            return embeddings.detach().cpu().numpy()



from typing import Any, Optional, Tuple

import nanopq
import numpy as np

from fast_forward.quantizer import Quantizer, QuantizerAttributes, QuantizerData


class NanoPQ(Quantizer):
    """Product quantizer that uses the [nanopq library](https://nanopq.readthedocs.io/en/stable/index.html).

    More information is available [here](https://nanopq.readthedocs.io/en/stable/source/api.html#nanopq.PQ).
    """

    def __init__(
        self, M: int, Ks: int, metric: str = "dot", verbose: bool = False
    ) -> None:
        """Instantiate a nanopq quantizer.

        Args:
            M (int): The number of subspaces.
            Ks (int): The number of codewords per subspace.
            metric (str, optional): The metric to use. Defaults to "dot".
            verbose (bool, optional): Enable verbosity. Defaults to False.
        """
        self._pq = nanopq.PQ(M=M, Ks=Ks, metric=metric, verbose=verbose)
        super().__init__()

    def _fit(self, vectors: np.ndarray, **kwargs: Any) -> None:
        self._pq.fit(vecs=vectors, **kwargs)

    @property
    def dtype(self) -> np.dtype:
        return self._pq.code_dtype

    @property
    def dims(self) -> Tuple[Optional[int], Optional[int]]:
        if self._pq.Ds is None:
            return None, self._pq.M
        return self._pq.Ds * self._pq.M, self._pq.M

    def _encode(self, vectors: np.ndarray) -> np.ndarray:
        return self._pq.encode(vecs=vectors)

    def _decode(self, codes: np.ndarray) -> np.ndarray:
        return self._pq.decode(codes=codes)

    def _get_state(self) -> Tuple[QuantizerAttributes, QuantizerData]:
        attributes, data = {}, {}

        for a in ("M", "Ks", "Ds", "metric", "verbose"):
            attributes[a] = getattr(self._pq, a)

        if (
            hasattr(self._pq, "codewords")
            and getattr(self._pq, "codewords") is not None
        ):
            data["codewords"] = self._pq.codewords

        return attributes, data

    @classmethod
    def _from_state(
        cls, attributes: QuantizerAttributes, data: QuantizerData
    ) -> "NanoPQ":
        quantizer = cls(
            M=attributes["M"],
            Ks=attributes["Ks"],
            metric=attributes["metric"],
            verbose=attributes["verbose"],
        )
        if "Ds" in attributes:
            quantizer._pq.Ds = attributes["Ds"]
        if "codewords" in data:
            quantizer._pq.codewords = data["codewords"]
        return quantizer


class NanoOPQ(Quantizer):
    """Optimized product quantizer that uses the [nanopq library](https://nanopq.readthedocs.io/en/stable/index.html).

    More information is available [here](https://nanopq.readthedocs.io/en/stable/source/api.html#nanopq.OPQ).
    """

    def __init__(
        self, M: int, Ks: int, metric: str = "dot", verbose: bool = False
    ) -> None:
        """Instantiate a nanopq optimized product quantizer.

        Args:
            M (int): The number of subspaces.
            Ks (int): The number of codewords per subspace.
            metric (str, optional): The metric to use. Defaults to "dot".
            verbose (bool, optional): Enable verbosity. Defaults to False.
        """
        self._opq = nanopq.OPQ(M=M, Ks=Ks, metric=metric, verbose=verbose)
        super().__init__()

    def _fit(self, vectors: np.ndarray, **kwargs: Any) -> None:
        self._opq.fit(vecs=vectors, **kwargs)

    @property
    def dtype(self) -> np.dtype:
        return self._opq.code_dtype

    @property
    def dims(self) -> Tuple[Optional[int], Optional[int]]:
        if self._opq.pq.Ds is None:
            return None, self._opq.pq.M
        return self._opq.pq.Ds * self._opq.pq.M, self._opq.pq.M

    def _encode(self, vectors: np.ndarray) -> np.ndarray:
        return self._opq.encode(vecs=vectors)

    def _decode(self, codes: np.ndarray) -> np.ndarray:
        return self._opq.decode(codes=codes)

    def _get_state(self) -> Tuple[QuantizerAttributes, QuantizerData]:
        attributes, data = {}, {}

        for a in ("M", "Ks", "Ds", "metric"):
            attributes[a] = getattr(self._opq.pq, a)

        attributes["verbose"] = self._opq.verbose

        if (
            hasattr(self._opq.pq, "codewords")
            and getattr(self._opq.pq, "codewords") is not None
        ):
            data["codewords"] = self._opq.pq.codewords

        if hasattr(self._opq, "R") and getattr(self._opq, "R") is not None:
            data["R"] = self._opq.R

        return attributes, data

    @classmethod
    def _from_state(
        cls, attributes: QuantizerAttributes, data: QuantizerData
    ) -> "NanoOPQ":
        quantizer = cls(
            M=attributes["M"],
            Ks=attributes["Ks"],
            metric=attributes["metric"],
            verbose=attributes["verbose"],
        )
        if "Ds" in attributes:
            quantizer._opq.pq.Ds = attributes["Ds"]
        if "codewords" in data:
            quantizer._opq.pq.codewords = data["codewords"]
        if "R" in data:
            quantizer._opq.R = data["R"]
        return quantizer



"""
.. include:: ../docs/util.md
"""

from typing import Callable

import numpy as np
import pandas as pd
from tqdm import tqdm

from fast_forward.index import Index, Mode
from fast_forward.ranking import Ranking


def to_ir_measures(ranking: Ranking) -> pd.DataFrame:
    """Return a ranking as a data frame suitable for the ir-measures library.

    Args:
        ranking (Ranking): The input ranking.

    Returns:
        pd.DataFrame: The data frame.
    """
    return ranking._df[["q_id", "id", "score"]].rename(
        columns={"q_id": "query_id", "id": "doc_id"}
    )


def cos_dist(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Cosine distance of two vectors.

    Args:
        a (np.ndarray): First vector.
        b (np.ndarray): Second vector.

    Returns:
        np.ndarray: Cosine distance.
    """
    return 1 - np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))


def create_coalesced_index(
    source_index: Index,
    target_index: Index,
    delta: float,
    distance_function: Callable[[np.ndarray, np.ndarray], float] = cos_dist,
    buffer_size: int = None,
) -> None:
    """Create a compressed index using sequential coalescing.

    Args:
        source_index (Index): The source index. Should contain multiple vectors for each document.
        target_index (Index): The target index. Must be empty.
        delta (float): The coalescing threshold.
        distance_function (Callable[[np.ndarray, np.ndarray], float]): The distance function. Defaults to cos_dist.
        buffer_size (int, optional): Use a buffer instead of adding all vectors at the end. Defaults to None.
    """
    assert len(target_index.doc_ids) == 0
    buffer_size = buffer_size or len(source_index.doc_ids)
    source_index.mode = Mode.MAXP

    def _coalesce(P):
        P_new = []
        A = []
        A_avg = None
        first_iteration = True
        for v in P:
            if first_iteration:
                first_iteration = False
            elif distance_function(v, A_avg) >= delta:
                P_new.append(A_avg)
                A = []
            A.append(v)
            A_avg = np.mean(A, axis=0)
        P_new.append(A_avg)
        return P_new

    vectors, doc_ids = [], []
    for doc_id in tqdm(source_index.doc_ids):

        # check if buffer is full
        if len(vectors) == buffer_size:
            target_index.add(np.array(vectors), doc_ids=doc_ids)
            vectors, doc_ids = [], []

        v_old, _ = source_index._get_vectors([doc_id])
        v_new = _coalesce(v_old)
        vectors.extend(v_new)
        doc_ids.extend([doc_id] * len(v_new))

    if len(vectors) > 0:
        target_index.add(np.array(vectors), doc_ids=doc_ids)

    assert source_index.doc_ids == target_index.doc_ids



import logging
from collections import defaultdict
from typing import Iterable, Iterator, List, Optional, Sequence, Set, Tuple

import numpy as np
from tqdm import tqdm

from fast_forward.encoder import Encoder
from fast_forward.index import IDSequence, Index, Mode
from fast_forward.quantizer import Quantizer

LOGGER = logging.getLogger(__name__)


class InMemoryIndex(Index):
    """Fast-Forward index that is held entirely in memory."""

    def __init__(
        self,
        query_encoder: Encoder = None,
        quantizer: Quantizer = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
        init_size: int = 2**14,
        alloc_size: int = 2**14,
    ) -> None:
        """Create an index.

        Args:
            query_encoder (Encoder, optional): The query encoder to use. Defaults to None.
            quantizer (Quantizer, optional): The quantizer to use. Defaults to None.
            mode (Mode, optional): Ranking mode. Defaults to Mode.MAXP.
            encoder_batch_size (int, optional): Query encoder batch size. Defaults to 32.
            init_size (int, optional): Initial index size. Defaults to 2**14.
            alloc_size (int, optional): Size of shard allocated when index is full. Defaults to 2**14.
        """
        self._shards = []
        self._init_size = init_size
        self._alloc_size = alloc_size
        self._idx_in_cur_shard = 0
        self._doc_id_to_idx = defaultdict(list)
        self._psg_id_to_idx = {}

        super().__init__(
            query_encoder=query_encoder,
            quantizer=quantizer,
            mode=mode,
            encoder_batch_size=encoder_batch_size,
        )

    def __len__(self) -> int:
        # account for the fact that the first shard might be larger
        if len(self._shards) < 2:
            return self._idx_in_cur_shard
        else:
            return (
                self._shards[0].shape[0]
                + (len(self._shards) - 2) * self._alloc_size
                + self._idx_in_cur_shard
            )

    def _get_internal_dim(self) -> Optional[int]:
        if len(self._shards) > 0:
            return self._shards[0].shape[-1]
        return None

    def _add(
        self,
        vectors: np.ndarray,
        doc_ids: Sequence[Optional[str]],
        psg_ids: Sequence[Optional[str]],
    ) -> None:
        # if this is the first call to _add, no shards exist
        if len(self._shards) == 0:
            self._shards.append(
                np.zeros((self._init_size, vectors.shape[-1]), dtype=vectors.dtype)
            )

        # assign passage and document IDs
        for i, doc_id in enumerate(doc_ids, len(self)):
            if doc_id is not None:
                self._doc_id_to_idx[doc_id].append(i)

        for i, psg_id in enumerate(psg_ids, len(self)):
            if psg_id is None:
                continue
            if psg_id in self._psg_id_to_idx:
                raise RuntimeError(f"Passage ID {psg_id} already exists.")
            self._psg_id_to_idx[psg_id] = i

        # add vectors to shards
        added = 0
        num_vectors = vectors.shape[0]
        while added < num_vectors:
            cur_shard_size, dim = self._shards[-1].shape

            # if current shard is full, add a new one
            if self._idx_in_cur_shard == cur_shard_size:
                LOGGER.debug("adding new shard")
                self._shards.append(np.zeros((self._alloc_size, dim)))
                self._idx_in_cur_shard = 0
                cur_shard_size = self._alloc_size

            to_add = min(
                num_vectors - added,
                cur_shard_size,
                cur_shard_size - self._idx_in_cur_shard,
            )
            self._shards[-1][
                self._idx_in_cur_shard : self._idx_in_cur_shard + to_add
            ] = vectors[added : added + to_add]
            added += to_add
            self._idx_in_cur_shard += to_add

    def consolidate(self) -> None:
        """Combine all shards of the index in one contiguous section in the memory."""
        # combine all shards up to the last one entirely, and take only whats in use of the last one
        self._shards = [
            np.concatenate(
                self._shards[:-1] + [self._shards[-1][: self._idx_in_cur_shard]]
            )
        ]
        self._idx_in_cur_shard = self._shards[0].shape[0]

    @property
    def doc_ids(self) -> Set[str]:
        return set(self._doc_id_to_idx.keys())

    @property
    def psg_ids(self) -> Set[str]:
        return set(self._psg_id_to_idx.keys())

    def _index_shards(self, idx: int) -> Tuple[int, int]:
        """Given an index in `[0, N]`, where `N` is the total number of vectors,
        compute the corresponding shard index and index within that shard.

        Args:
            idx (int): The input index.

        Returns:
            Tuple[int, int]: Shard index and index within that shard.
        """
        # the first shard might be larger
        if idx < self._shards[0].shape[0]:
            shard_idx = 0
            idx_in_shard = idx
        else:
            idx -= self._shards[0].shape[0]
            shard_idx = int(idx / self._alloc_size) + 1
            idx_in_shard = idx % self._alloc_size
        return shard_idx, idx_in_shard

    def _get_vectors(self, ids: Iterable[str]) -> Tuple[np.ndarray, List[List[int]]]:
        items_by_shard = defaultdict(list)
        for id in ids:
            if self.mode in (Mode.MAXP, Mode.AVEP) and id in self._doc_id_to_idx:
                idxs = self._doc_id_to_idx[id]
            elif self.mode == Mode.FIRSTP and id in self._doc_id_to_idx:
                idxs = [self._doc_id_to_idx[id][0]]
            elif self.mode == Mode.PASSAGE and id in self._psg_id_to_idx:
                idxs = [self._psg_id_to_idx[id]]
            else:
                LOGGER.warning("no vectors for %s", id)
                idxs = []

            for idx in idxs:
                shard_idx, idx_in_shard = self._index_shards(idx)
                items_by_shard[shard_idx].append((idx_in_shard, id))

        result_vectors = []
        result_ids = defaultdict(list)
        items_so_far = 0
        for shard_idx, items in items_by_shard.items():
            idxs, ids_ = zip(*items)
            result_vectors.append(self._shards[shard_idx][list(idxs)])
            for i, id_in_shard in enumerate(ids_):
                result_ids[id_in_shard].append(i + items_so_far)
            items_so_far += len(items)

        if len(result_vectors) == 0:
            return np.array([]), []
        return np.concatenate(result_vectors), [result_ids[id] for id in ids]

    def _batch_iter(
        self, batch_size: int
    ) -> Iterator[Tuple[np.ndarray, IDSequence, IDSequence]]:
        LOGGER.info("creating ID mappings for this index")
        idx_to_doc_id = {
            idx: doc_id
            for doc_id, idxs in tqdm(self._doc_id_to_idx.items())
            for idx in idxs
        }
        idx_to_psg_id = {
            idx: psg_id for psg_id, idx in tqdm(self._psg_id_to_idx.items())
        }

        num_vectors = len(self)
        for i in range(0, num_vectors, batch_size):
            j = min(i + batch_size, num_vectors)

            # the current batch is between i (incl.) and j (excl.)
            i_shard_idx, i_idx_in_shard = self._index_shards(i)
            j_shard_idx, j_idx_in_shard = self._index_shards(j - 1)

            arrays = []

            # if the batch spans multiple shards, collect them in a list
            while i_shard_idx < j_shard_idx:
                arrays.append(self._shards[i_shard_idx][i_idx_in_shard:])
                i_shard_idx += 1
                i_idx_in_shard = 0

            # now i_shard_idx == j_shard_idx
            arrays.append(
                self._shards[i_shard_idx][i_idx_in_shard : j_idx_in_shard + 1]
            )

            yield (
                np.concatenate(arrays),
                list(map(idx_to_doc_id.get, range(i, j))),
                list(map(idx_to_psg_id.get, range(i, j))),
            )



import itertools
import shutil
import tempfile
import unittest
from pathlib import Path

import numpy as np
import pandas as pd

from fast_forward import InMemoryIndex, Mode, OnDiskIndex, Ranking
from fast_forward.encoder import LambdaEncoder
from fast_forward.quantizer.nanopq import NanoPQ
from fast_forward.util import create_coalesced_index

DUMMY_QUERIES = {"q1": "query 1", "q2": "query 2"}
DUMMY_DOC_IDS = ["d0", "d0", "d1", "d2", "d3"]
UNIQUE_DUMMY_DOC_IDS = list(set(DUMMY_DOC_IDS))
DUMMY_PSG_IDS = ["p0", "p1", "p2", "p3", "p4"]
DUMMY_VECTORS = np.array(
    [
        [1, 0, 0, 0, 0],
        [1, 1, 0, 0, 0],
        [1, 1, 1, 0, 0],
        [1, 1, 1, 1, 0],
        [1, 1, 1, 1, 1],
    ]
)
DUMMY_NUM, DUMMY_DIM = DUMMY_VECTORS.shape
DUMMY_DOC_RUN = {
    "q1": {"d0": 100, "d1": 2, "d2": 3, "d3": 200},
    "q2": {"d0": 400, "d1": 5, "d2": 6, "d3": 800, "dx": 7},
}
DUMMY_DOC_RANKING = Ranking.from_run(DUMMY_DOC_RUN, queries=DUMMY_QUERIES)
DUMMY_PSG_RUN = {
    "q1": {"p0": 100, "p1": 2, "p2": 3, "p3": 4, "p4": 5},
    "q2": {"p0": 500, "p1": 6, "p2": 7, "p3": 8, "p4": 9},
}
DUMMY_PSG_RANKING = Ranking.from_run(DUMMY_PSG_RUN, queries=DUMMY_QUERIES)
DUMMY_ENCODER = LambdaEncoder(lambda _: np.array([1, 1, 1, 1, 1]))

DUMMY_QUANTIZER = NanoPQ(2, 8)
DUMMY_QUANTIZER.fit(np.random.normal(size=(16, 16)).astype(np.float32))


class TestIndex(unittest.TestCase):
    __test__ = False

    def setUp(self):
        self.doc_psg_index.add(
            vectors=DUMMY_VECTORS, doc_ids=DUMMY_DOC_IDS, psg_ids=DUMMY_PSG_IDS
        )

        # some vectors have only a document ID, some have only a passage ID, some have both
        self.index_partial_ids.add(
            vectors=DUMMY_VECTORS,
            doc_ids=[None, None] + DUMMY_DOC_IDS[2:],
            psg_ids=DUMMY_PSG_IDS[:-2] + [None, None],
        )
        # vectors have only document IDs
        self.index_partial_ids.add(vectors=DUMMY_VECTORS[:2], doc_ids=DUMMY_DOC_IDS[:2])
        # vectors have only passage IDs
        self.index_partial_ids.add(
            vectors=DUMMY_VECTORS[-2:], psg_ids=DUMMY_PSG_IDS[-2:]
        )

        self.doc_index.add(vectors=DUMMY_VECTORS, doc_ids=DUMMY_DOC_IDS)
        self.psg_index.add(vectors=DUMMY_VECTORS, psg_ids=DUMMY_PSG_IDS)

        self.quantized_index.add(
            vectors=np.random.normal(size=(5, DUMMY_QUANTIZER.dims[0])).astype(
                np.float32
            ),
            doc_ids=DUMMY_DOC_IDS,
        )

    def test_properties(self):
        self.assertEqual(set(DUMMY_DOC_IDS), self.doc_psg_index.doc_ids)
        self.assertEqual(set(DUMMY_PSG_IDS), self.doc_psg_index.psg_ids)
        self.assertEqual(DUMMY_NUM, len(self.doc_psg_index))
        self.assertEqual(DUMMY_DIM, self.doc_psg_index.dim)

        self.assertEqual(set(DUMMY_DOC_IDS), self.index_partial_ids.doc_ids)
        self.assertEqual(set(DUMMY_PSG_IDS), self.index_partial_ids.psg_ids)
        self.assertEqual(DUMMY_NUM + 4, len(self.index_partial_ids))
        self.assertEqual(DUMMY_DIM, self.index_partial_ids.dim)

        self.assertEqual(set(DUMMY_DOC_IDS), self.doc_index.doc_ids)
        self.assertEqual(0, len(self.doc_index.psg_ids))
        self.assertEqual(DUMMY_NUM, len(self.doc_index))
        self.assertEqual(DUMMY_DIM, self.doc_index.dim)

        self.assertEqual(set(DUMMY_PSG_IDS), self.psg_index.psg_ids)
        self.assertEqual(0, len(self.psg_index.doc_ids))
        self.assertEqual(DUMMY_NUM, len(self.psg_index))
        self.assertEqual(DUMMY_DIM, self.psg_index.dim)

        self.assertEqual(16, self.quantized_index.dim)

    def test_add_retrieve(self):
        self.assertEqual(0, len(self.index))

        data = np.random.normal(size=(80, 16))
        doc_ids = [f"doc_{int(i/2)}" for i in range(data.shape[0])]
        psg_ids = [f"psg_{i}" for i in range(data.shape[0])]

        # successively add parts of the data and make sure we still get the correct vectors and indices back as the index grows
        for lower, upper in [(0, 8), (8, 24), (24, 80)]:
            self.index.add(
                data[lower:upper],
                doc_ids=doc_ids[lower:upper],
                psg_ids=psg_ids[lower:upper],
            )
            self.assertEqual(upper, len(self.index))

            self.index.mode = Mode.PASSAGE
            vecs, idxs = self.index._get_vectors(psg_ids[lower:upper])
            _test_vectors(
                vecs, idxs, data[lower:upper], [[idx] for idx in range(upper - lower)]
            )

            self.index.mode = Mode.MAXP
            vecs, idxs = self.index._get_vectors(
                [f"doc_{i}" for i in range(int(lower / 2), int(upper / 2))]
            )
            np.testing.assert_almost_equal(vecs, data[lower:upper], decimal=6)
            self.assertEqual(
                [[2 * idx, 2 * idx + 1] for idx in range(int((upper - lower) / 2))],
                idxs,
            )

    def test_maxp(self):
        self.doc_psg_index.mode = Mode.MAXP
        result = self.doc_psg_index(DUMMY_DOC_RANKING)
        self.assertEqual(
            result,
            Ranking.from_run(
                {
                    "q1": {"d0": 2, "d1": 3, "d2": 4, "d3": 5},
                    "q2": {"d0": 2, "d1": 3, "d2": 4, "d3": 5},
                }
            ),
        )

    def test_firstp(self):
        expected = Ranking.from_run(
            {
                "q1": {"d0": 1, "d1": 3, "d2": 4, "d3": 5},
                "q2": {"d0": 1, "d1": 3, "d2": 4, "d3": 5},
            }
        )
        self.doc_psg_index.mode = Mode.FIRSTP
        self.assertEqual(
            self.doc_psg_index(DUMMY_DOC_RANKING),
            expected,
        )
        self.index_partial_ids.mode = Mode.FIRSTP
        self.assertEqual(
            self.doc_psg_index(DUMMY_DOC_RANKING),
            expected,
        )

    def test_avep(self):
        expected = Ranking.from_run(
            {
                "q1": {"d0": 1.5, "d1": 3, "d2": 4, "d3": 5},
                "q2": {"d0": 1.5, "d1": 3, "d2": 4, "d3": 5},
            }
        )

        self.doc_psg_index.mode = Mode.AVEP
        self.assertEqual(
            self.doc_psg_index(DUMMY_DOC_RANKING),
            expected,
        )
        self.index_partial_ids.mode = Mode.AVEP
        self.assertEqual(
            self.index_partial_ids(DUMMY_DOC_RANKING),
            expected,
        )

    def test_passage(self):
        expected = Ranking.from_run(
            {
                "q1": {"p0": 1, "p1": 2, "p2": 3, "p3": 4, "p4": 5},
                "q2": {"p0": 1, "p1": 2, "p2": 3, "p3": 4, "p4": 5},
            }
        )
        self.doc_psg_index.mode = Mode.PASSAGE
        self.assertEqual(
            self.doc_psg_index(DUMMY_PSG_RANKING),
            expected,
        )
        self.index_partial_ids.mode = Mode.PASSAGE
        self.assertEqual(
            self.index_partial_ids(DUMMY_PSG_RANKING),
            expected,
        )

    def test_errors(self):
        # no IDs
        with self.assertRaises(ValueError):
            self.index_no_enc.add(DUMMY_VECTORS, doc_ids=None, psg_ids=None)

        # too few IDs
        with self.assertRaises(ValueError):
            self.index_no_enc.add(
                DUMMY_VECTORS, doc_ids=DUMMY_DOC_IDS[:-2], psg_ids=None
            )
        with self.assertRaises(ValueError):
            self.index_no_enc.add(
                DUMMY_VECTORS, doc_ids=None, psg_ids=DUMMY_PSG_IDS[:-2]
            )

        # missing ID
        with self.assertRaises(ValueError):
            self.index_no_enc.add(
                DUMMY_VECTORS,
                doc_ids=[None] + DUMMY_DOC_IDS[1:],
                psg_ids=[None] + DUMMY_PSG_IDS[1:],
            )

        # encoding without encoder
        with self.assertRaises(RuntimeError):
            self.index_no_enc.encode_queries(["test"])

        # adding vectors with wrong dimension
        self.index_wrong_dim.add(np.array([[0, 0], [1, 1]]), doc_ids=["d1", "d2"])
        with self.assertRaises(ValueError):
            self.index_wrong_dim.add(
                np.array([[0, 0, 0], [1, 1, 1]]), doc_ids=["d3", "d4"]
            )

        # ranking without queries
        ranking_no_queries = Ranking.from_run(DUMMY_DOC_RUN)
        with self.assertRaises(ValueError):
            self.doc_psg_index(ranking_no_queries)

        # early stopping without required parameters
        with self.assertRaises(ValueError):
            self.doc_psg_index(
                DUMMY_DOC_RANKING, early_stopping=10, early_stopping_alpha=None
            )
        with self.assertRaises(ValueError):
            self.doc_psg_index(
                DUMMY_DOC_RANKING, early_stopping=10, early_stopping_intervals=None
            )

        # adding a quantizer to an index that's not empty
        with self.assertRaises(RuntimeError):
            self.doc_psg_index.quantizer = DUMMY_QUANTIZER

    def test_early_stopping(self):
        self.early_stopping_index.add(
            np.stack([[1, 0], [1, 1]] * 10), psg_ids=[f"p{i}" for i in range(20)]
        )
        r = Ranking(
            pd.DataFrame(
                [
                    {"q_id": q, "query": q, "id": f"p{i}", "score": i}
                    for i in range(20)
                    for q in ("q1", "q2")
                ]
            )
        )

        result_expected = Ranking(
            pd.DataFrame(
                [
                    {"q_id": "q2", "id": "p19", "score": 20.0},
                    {"q_id": "q2", "id": "p17", "score": 20.0},
                    {"q_id": "q2", "id": "p15", "score": 20.0},
                    {"q_id": "q2", "id": "p13", "score": 20.0},
                    {"q_id": "q2", "id": "p11", "score": 20.0},
                    {"q_id": "q2", "id": "p18", "score": 10.0},
                    {"q_id": "q2", "id": "p16", "score": 10.0},
                    {"q_id": "q2", "id": "p14", "score": 10.0},
                    {"q_id": "q2", "id": "p12", "score": 10.0},
                    {"q_id": "q2", "id": "p10", "score": 10.0},
                    {"q_id": "q1", "id": "p19", "score": 20.0},
                    {"q_id": "q1", "id": "p17", "score": 20.0},
                    {"q_id": "q1", "id": "p15", "score": 20.0},
                    {"q_id": "q1", "id": "p13", "score": 20.0},
                    {"q_id": "q1", "id": "p11", "score": 20.0},
                    {"q_id": "q1", "id": "p18", "score": 10.0},
                    {"q_id": "q1", "id": "p16", "score": 10.0},
                    {"q_id": "q1", "id": "p14", "score": 10.0},
                    {"q_id": "q1", "id": "p12", "score": 10.0},
                    {"q_id": "q1", "id": "p10", "score": 10.0},
                ]
            )
        )

        self.assertEqual(
            self.early_stopping_index(
                r,
                early_stopping=5,
                early_stopping_alpha=0.5,
                early_stopping_intervals=(2, 5, 10, 20),
            ),
            result_expected,
        )

        # order of intervals should make no difference
        self.assertEqual(
            self.early_stopping_index(
                r,
                early_stopping=5,
                early_stopping_alpha=0.5,
                early_stopping_intervals=(5, 2, 20, 10),
            ),
            result_expected,
        )

    def test_coalescing(self):
        # delta = 0.3: vectors of d0 should be averaged
        create_coalesced_index(self.doc_index, self.coalesced_indexes[0], 0.3)
        self.assertEqual(self.doc_index.doc_ids, self.coalesced_indexes[0].doc_ids)
        d0_vector_expected = np.average([DUMMY_VECTORS[0], DUMMY_VECTORS[1]], axis=0)
        d0_vectors, _ = self.coalesced_indexes[0]._get_vectors(["d0"])
        self.assertEqual(1, len(d0_vectors))
        self.assertTrue(np.array_equal(d0_vector_expected, d0_vectors[0]))

        # delta = 0.2: nothing should change
        create_coalesced_index(
            self.doc_index, self.coalesced_indexes[1], 0.2, buffer_size=2
        )
        self.assertEqual(self.doc_index.doc_ids, self.coalesced_indexes[1].doc_ids)
        for doc_id in self.doc_index.doc_ids:
            vectors_1, _ = self.doc_index._get_vectors([doc_id])
            vectors_2, _ = self.coalesced_indexes[1]._get_vectors([doc_id])
            self.assertEqual(len(vectors_1), len(vectors_2))
            for v1, v2 in zip(vectors_1, vectors_2):
                self.assertTrue(np.array_equal(v1, v2))

    def test_iter(self):
        for index in self.iter_indexes:
            index.add(DUMMY_VECTORS, doc_ids=DUMMY_DOC_IDS, psg_ids=DUMMY_PSG_IDS)
            for batch_size in (1, 3, 5, 10):
                vectors, doc_ids, psg_ids = zip(*index.batch_iter(batch_size))
                np.testing.assert_equal(DUMMY_VECTORS, np.concatenate(vectors))
                self.assertEqual(
                    DUMMY_DOC_IDS, list(itertools.chain.from_iterable(doc_ids))
                )
                self.assertEqual(
                    DUMMY_PSG_IDS, list(itertools.chain.from_iterable(psg_ids))
                )

    def test_quantization(self):
        self.assertEqual(2, self.quantized_index._get_internal_dim())

        # make sure the dimensions of the returned vetors match the original dimension
        for vec, _, _ in self.quantized_index:
            self.assertEqual(16, vec.shape[0])

        self.quantized_index.mode = Mode.MAXP
        self.assertEqual(
            self.quantized_index._get_vectors(UNIQUE_DUMMY_DOC_IDS)[0].shape, (5, 2)
        )


class TestInMemoryIndex(TestIndex):
    __test__ = True

    def setUp(self):
        self.index = InMemoryIndex(init_size=32, alloc_size=32)
        self.doc_psg_index = InMemoryIndex(DUMMY_ENCODER)
        self.index_partial_ids = InMemoryIndex(DUMMY_ENCODER)
        self.doc_index = InMemoryIndex(DUMMY_ENCODER)
        self.psg_index = InMemoryIndex(DUMMY_ENCODER)
        self.index_no_enc = InMemoryIndex(query_encoder=None)
        self.index_wrong_dim = InMemoryIndex(query_encoder=None)
        self.early_stopping_index = InMemoryIndex(
            LambdaEncoder(lambda q: np.array([10, 10])), mode=Mode.PASSAGE
        )
        self.coalesced_indexes = [
            InMemoryIndex(mode=Mode.MAXP),
            InMemoryIndex(mode=Mode.MAXP),
        ]
        self.iter_indexes = [
            InMemoryIndex(init_size=2, alloc_size=2),
            InMemoryIndex(init_size=5),
        ]
        self.quantized_index = InMemoryIndex(quantizer=DUMMY_QUANTIZER)
        super().setUp()

    def test_consolidate(self):
        index = InMemoryIndex(init_size=8, alloc_size=4)
        data = data = np.random.normal(size=(32, 16))
        psg_ids = [f"psg_{i}" for i in range(32)]

        index.add(data[:14], psg_ids=psg_ids[:14])
        index.consolidate()
        vecs, idxs = index._get_vectors(psg_ids[:14])
        _test_vectors(vecs, idxs, data[:14], [[idx] for idx in range(14)])

        index.add(data[14:32], psg_ids=psg_ids[14:32])
        index.consolidate()
        vecs, idxs = index._get_vectors(psg_ids)
        _test_vectors(vecs, idxs, data, [[idx] for idx in range(32)])


class TestOnDiskIndex(TestIndex):
    __test__ = True

    def setUp(self):
        self.temp_dir = Path(tempfile.mkdtemp())
        self.index = OnDiskIndex(
            self.temp_dir / "index.h5", init_size=32, resize_min_val=32
        )
        self.doc_psg_index = OnDiskIndex(
            self.temp_dir / "doc_psg_index.h5",
            DUMMY_ENCODER,
        )
        self.index_partial_ids = OnDiskIndex(
            self.temp_dir / "index_partial_ids.h5",
            DUMMY_ENCODER,
        )
        self.doc_index = OnDiskIndex(
            self.temp_dir / "doc_index.h5",
            DUMMY_ENCODER,
        )
        self.psg_index = OnDiskIndex(
            self.temp_dir / "psg_index.h5",
            DUMMY_ENCODER,
        )
        self.index_no_enc = OnDiskIndex(
            self.temp_dir / "index_no_enc.h5", query_encoder=None
        )
        self.index_wrong_dim = OnDiskIndex(
            self.temp_dir / "index_wrong_dim.h5", query_encoder=None
        )
        self.early_stopping_index = OnDiskIndex(
            self.temp_dir / "early_stopping_index.h5",
            LambdaEncoder(lambda q: np.array([10, 10])),
            mode=Mode.PASSAGE,
        )
        self.coalesced_indexes = [
            OnDiskIndex(self.temp_dir / "coalesced_index_1.h5", mode=Mode.MAXP),
            OnDiskIndex(self.temp_dir / "coalesced_index_2.h5", mode=Mode.MAXP),
        ]
        self.iter_indexes = [
            OnDiskIndex(
                self.temp_dir / "iter_index_1.h5",
                init_size=2,
                resize_min_val=2,
            ),
            OnDiskIndex(self.temp_dir / "iter_index_2.h5", init_size=5),
        ]
        self.quantized_index = OnDiskIndex(
            self.temp_dir / "quantized_index.h5", quantizer=DUMMY_QUANTIZER
        )
        super().setUp()

    def test_load(self):
        # test whether vectors are preserved properly
        shutil.copy(
            self.temp_dir / "doc_psg_index.h5", self.temp_dir / "doc_psg_index_copy.h5"
        )
        index_copied = OnDiskIndex.load(self.temp_dir / "doc_psg_index_copy.h5")
        self.assertEqual(index_copied.doc_ids, self.doc_psg_index.doc_ids)
        self.assertEqual(index_copied.psg_ids, self.doc_psg_index.psg_ids)
        self.doc_psg_index.mode = Mode.PASSAGE
        index_copied.mode = Mode.PASSAGE
        _test_get_vectors(index_copied, self.doc_psg_index, DUMMY_PSG_IDS)
        self.doc_psg_index.mode = Mode.MAXP
        index_copied.mode = Mode.MAXP
        _test_get_vectors(index_copied, self.doc_psg_index, UNIQUE_DUMMY_DOC_IDS)

        shutil.copy(self.temp_dir / "doc_index.h5", self.temp_dir / "doc_index_copy.h5")
        index_copied = OnDiskIndex.load(self.temp_dir / "doc_index_copy.h5")
        self.assertEqual(index_copied.doc_ids, self.doc_index.doc_ids)
        self.assertEqual(index_copied.psg_ids, self.doc_index.psg_ids)
        self.doc_index.mode = Mode.MAXP
        index_copied.mode = Mode.MAXP
        _test_get_vectors(index_copied, self.doc_index, UNIQUE_DUMMY_DOC_IDS)

        shutil.copy(self.temp_dir / "psg_index.h5", self.temp_dir / "psg_index_copy.h5")
        index_copied = OnDiskIndex.load(self.temp_dir / "psg_index_copy.h5")
        self.assertEqual(index_copied.doc_ids, self.psg_index.doc_ids)
        self.assertEqual(index_copied.psg_ids, self.psg_index.psg_ids)
        self.psg_index.mode = Mode.PASSAGE
        index_copied.mode = Mode.PASSAGE
        _test_get_vectors(index_copied, self.psg_index, DUMMY_PSG_IDS)

        # test whether quantizers are loaded properly
        shutil.copy(
            self.temp_dir / "quantized_index.h5",
            self.temp_dir / "quantized_index_copy.h5",
        )
        quantized_index_copied = OnDiskIndex.load(
            self.temp_dir / "quantized_index_copy.h5"
        )
        self.assertEqual(
            quantized_index_copied.quantizer, self.quantized_index.quantizer
        )
        self.quantized_index.mode = Mode.PASSAGE
        quantized_index_copied.mode = Mode.PASSAGE
        _test_get_vectors(index_copied, self.psg_index, DUMMY_PSG_IDS)

        # test loading an empty index
        OnDiskIndex(self.temp_dir / "empty_index.h5")
        empty_index_loaded = OnDiskIndex.load(self.temp_dir / "empty_index.h5")
        self.assertEqual(len(empty_index_loaded.doc_ids), 0)
        self.assertEqual(len(empty_index_loaded.doc_ids), 0)

    def test_store_quantizer(self):
        # create index with quantizer and then replace the quantizer
        index_with_quantizer = OnDiskIndex(self.temp_dir / "index_with_quantizer.h5")
        index_with_quantizer.quantizer = DUMMY_QUANTIZER
        new_quantizer = NanoPQ(2, 8)
        new_quantizer.fit(np.random.normal(size=(16, 16)).astype(np.float32))
        index_with_quantizer.quantizer = new_quantizer

        # load index again and check whether the new quantizer has been stored
        del index_with_quantizer
        index_with_quantizer = OnDiskIndex.load(
            self.temp_dir / "index_with_quantizer.h5"
        )
        self.assertEqual(new_quantizer, index_with_quantizer.quantizer)

    def test_to_memory(self):
        for index, params in [
            (self.doc_index, [(Mode.MAXP, UNIQUE_DUMMY_DOC_IDS)]),
            (self.psg_index, [(Mode.PASSAGE, DUMMY_PSG_IDS)]),
            (
                self.doc_psg_index,
                [(Mode.MAXP, UNIQUE_DUMMY_DOC_IDS), (Mode.PASSAGE, DUMMY_PSG_IDS)],
            ),
        ]:
            mem_index = index.to_memory()
            mem_index_buffered = index.to_memory(buffer_size=2)

            for mode, ids in params:
                index.mode = mode
                mem_index.mode = mode
                mem_index_buffered.mode = mode

                self.assertEqual(mem_index.doc_ids, index.doc_ids)
                self.assertEqual(mem_index.psg_ids, index.psg_ids)
                self.assertEqual(mem_index_buffered.doc_ids, index.doc_ids)
                self.assertEqual(mem_index_buffered.psg_ids, index.psg_ids)

                _test_get_vectors(mem_index, index, ids)
                _test_get_vectors(mem_index_buffered, index, ids)

        mem_quantized_index = self.quantized_index.to_memory()
        self.assertEqual(
            mem_quantized_index._quantizer._pq, self.quantized_index._quantizer._pq
        )

    def test_max_id_length(self):
        index = OnDiskIndex(self.temp_dir / "max_id_length_index.h5", max_id_length=3)
        vectors = np.zeros(shape=(16, 16))
        doc_ids_ok = ["d1"] * 16
        psg_ids_ok = [f"p{i}" for i in range(16)]
        index.add(vectors, doc_ids=doc_ids_ok, psg_ids=psg_ids_ok)

        doc_ids_long = [doc_id + "-long" for doc_id in doc_ids_ok]
        psg_ids_long = [psg_id + "-long" for psg_id in psg_ids_ok]

        with self.assertRaises(RuntimeError):
            index.add(vectors, doc_ids=doc_ids_long)
        with self.assertRaises(RuntimeError):
            index.add(vectors, psg_ids=psg_ids_long)

        # make sure the index remains unchanged in case of the error
        self.assertEqual(index.doc_ids, set(doc_ids_ok))
        self.assertEqual(index.psg_ids, set(psg_ids_ok))
        self.assertEqual(16, len(index))

    def test_ds_buffer_size(self):
        index = OnDiskIndex(
            self.temp_dir / "ds_buffer_size_index.h5",
            mode=Mode.PASSAGE,
            ds_buffer_size=5,
        )
        psg_reps = np.random.normal(size=(16, 16))
        psg_ids = [f"p{i}" for i in range(16)]
        index.add(psg_reps, psg_ids=psg_ids)
        vecs, id_idxs = index._get_vectors(psg_ids)
        np.testing.assert_almost_equal(
            vecs[id_idxs], psg_reps.reshape((16, 1, 16)), decimal=6
        )

    def tearDown(self):
        shutil.rmtree(self.temp_dir)


def _test_get_vectors(index_1, index_2, ids):
    vecs_1, idxs_1 = index_1._get_vectors(ids)
    vecs_2, idxs_2 = index_2._get_vectors(ids)
    _test_vectors(vecs_1, idxs_1, vecs_2, idxs_2)


def _test_vectors(vecs_1, idxs_1, vecs_2, idxs_2):
    # this accounts for the fact that the indices returned by _get_vectors may be different,
    # but the overall result is still the same
    for i, j in zip(idxs_1, idxs_2):
        np.testing.assert_almost_equal(vecs_1[i], vecs_2[j], decimal=6)


if __name__ == "__main__":
    unittest.main()



"""
.. include:: ../docs/index.md
"""

import abc
import logging
from enum import Enum
from time import perf_counter
from typing import Iterable, Iterator, List, Optional, Sequence, Set, Tuple

import numpy as np
import pandas as pd

from fast_forward.encoder import Encoder
from fast_forward.quantizer import Quantizer
from fast_forward.ranking import Ranking

LOGGER = logging.getLogger(__name__)


class Mode(Enum):
    """Enum used to set the ranking mode of an index."""

    PASSAGE = 1
    MAXP = 2
    FIRSTP = 3
    AVEP = 4


IDSequence = Sequence[Optional[str]]


class Index(abc.ABC):
    """Abstract base class for Fast-Forward indexes."""

    _query_encoder: Encoder = None
    _quantizer: Quantizer = None

    def __init__(
        self,
        query_encoder: Encoder = None,
        quantizer: Quantizer = None,
        mode: Mode = Mode.MAXP,
        encoder_batch_size: int = 32,
    ) -> None:
        """Create an index.

        Args:
            query_encoder (Encoder, optional): The query encoder to use. Defaults to None.
            quantizer (Quantizer, optional): The quantizer to use. Defaults to None.
            mode (Mode, optional): Ranking mode. Defaults to Mode.MAXP.
            encoder_batch_size (int, optional): Encoder batch size. Defaults to 32.
        """
        super().__init__()
        if query_encoder is not None:
            self.query_encoder = query_encoder
        self.mode = mode
        if quantizer is not None:
            self.quantizer = quantizer
        self._encoder_batch_size = encoder_batch_size

    def encode_queries(self, queries: Sequence[str]) -> np.ndarray:
        """Encode queries.

        Args:
            queries (Sequence[str]): The queries to encode.

        Raises:
            RuntimeError: When no query encoder exists.

        Returns:
            np.ndarray: The query representations.
        """
        if self.query_encoder is None:
            raise RuntimeError("Index does not have a query encoder.")

        result = []
        for i in range(0, len(queries), self._encoder_batch_size):
            batch = queries[i : i + self._encoder_batch_size]
            result.append(self.query_encoder(batch))
        return np.concatenate(result)

    @property
    def query_encoder(self) -> Optional[Encoder]:
        """Return the query encoder if it exists.

        Returns:
            Optional[Encoder]: The query encoder (if any).
        """
        return self._query_encoder

    @query_encoder.setter
    def query_encoder(self, encoder: Encoder) -> None:
        """Set the query encoder.

        Args:
            encoder (Encoder): The new query encoder.
        """
        assert isinstance(encoder, Encoder)
        self._query_encoder = encoder

    @property
    def quantizer(self) -> Optional[Quantizer]:
        """Return the quantizer if it exists.

        Returns:
            Optional[Quantizer]: The quantizer (if any).
        """
        return self._quantizer

    @quantizer.setter
    def quantizer(self, quantizer: Quantizer) -> None:
        """Set the quantizer. This is only possible before any vectors are added to the index.

        Raises:
            RuntimeError: When the index is not empty.

        Args:
            quantizer (Quantizer): The new quantizer.
        """
        assert isinstance(quantizer, Quantizer)

        if len(self) > 0:
            raise RuntimeError("Quantizers can only be attached to empty indexes.")
        self._quantizer = quantizer
        quantizer.set_attached()

    @property
    def mode(self) -> Mode:
        """Return the ranking mode.

        Returns:
            Mode: The ranking mode.
        """
        return self._mode

    @mode.setter
    def mode(self, mode: Mode) -> None:
        """Set the ranking mode.

        Args:
            mode (Mode): The new ranking mode.
        """
        assert isinstance(mode, Mode)
        self._mode = mode

    @abc.abstractmethod
    def _get_internal_dim(self) -> Optional[int]:
        """Return the dimensionality of the vectors (or codes) in the index (internal method).

        If no vectors exist, return None. If a quantizer is used, return the dimension of the codes.

        Returns:
            Optional[int]: The dimensionality (if any).
        """
        pass

    @property
    def dim(self) -> Optional[int]:
        """Return the dimensionality of the vector index.

        May return None if there are no vectors.

        If a quantizer is used, the dimension before quantization is returned.

        Returns:
            Optional[int]: The dimensionality (if any).
        """
        if self._quantizer is not None:
            return self._quantizer.dims[0]
        return self._get_internal_dim()

    @property
    @abc.abstractmethod
    def doc_ids(self) -> Set[str]:
        """Return all unique document IDs.

        Returns:
            Set[str]: The document IDs.
        """
        pass

    @property
    @abc.abstractmethod
    def psg_ids(self) -> Set[str]:
        """Return all unique passage IDs.

        Returns:
            Set[str]: The passage IDs.
        """
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        """The number of vectors in the index.

        Returns:
            int: The index size.
        """
        pass

    @abc.abstractmethod
    def _add(
        self,
        vectors: np.ndarray,
        doc_ids: Sequence[Optional[str]],
        psg_ids: Sequence[Optional[str]],
    ) -> None:
        """Add vector representations and corresponding IDs to the index.

        Document IDs may have duplicates, passage IDs are assumed to be unique. Vectors may be quantized.

        Args:
            vectors (np.ndarray): The representations, shape `(num_vectors, dim)` or `(num_vectors, quantized_dim)`.
            doc_ids (Sequence[Optional[str]]): The corresponding document IDs.
            psg_ids (Sequence[Optional[str]]): The corresponding passage IDs.
        """
        pass

    def add(
        self,
        vectors: np.ndarray,
        doc_ids: Sequence[Optional[str]] = None,
        psg_ids: Sequence[Optional[str]] = None,
    ) -> None:
        """Add vector representations and corresponding IDs to the index.

        Only one of `doc_ids` and `psg_ids` may be None. Individual IDs in the sequence may also be None,
        but each vector must have at least one associated ID.

        Document IDs may have duplicates, passage IDs must be unique.

        Args:
            vectors (np.ndarray): The representations, shape `(num_vectors, dim)`.
            doc_id (Sequence[Optional[str]], optional): The corresponding document IDs (may be duplicate). Defaults to None.
            psg_id (Sequence[Optional[str]], optional): The corresponding passage IDs (must be unique). Defaults to None.

        Raises:
            ValueError: When there are no document IDs and no passage IDs.
            ValueError: When the number of IDs does not match the number of vectors.
            ValueError: When the input vector and index dimensionalities don't match.
            ValueError: When a vector has neither a document nor a passage ID.
            RuntimeError: When items can't be added to the index for any reason.
        """
        if doc_ids is None and psg_ids is None:
            raise ValueError("At least one of doc_ids and psg_ids must be provided.")

        num_vectors, dim = vectors.shape

        if doc_ids is None:
            doc_ids = [None] * num_vectors
        if psg_ids is None:
            psg_ids = [None] * num_vectors
        if not len(doc_ids) == len(psg_ids) == num_vectors:
            raise ValueError("Number of IDs does not match number of vectors.")

        if self.dim is not None and dim != self.dim:
            raise ValueError(
                f"Input vector dimensionality ({dim}) does not match index dimensionality ({self.dim})."
            )

        for doc_id, psg_id in zip(doc_ids, psg_ids):
            if doc_id is None and psg_id is None:
                raise ValueError("Vector has neither document nor passage ID.")

        self._add(
            vectors if self.quantizer is None else self.quantizer.encode(vectors),
            doc_ids,
            psg_ids,
        )

    @abc.abstractmethod
    def _get_vectors(self, ids: Iterable[str]) -> Tuple[np.ndarray, List[List[int]]]:
        """Return:
            * A single array containing all vectors necessary to compute the scores for each document/passage.
            * For each document/passage (in the same order as the IDs), a list of integers (depending on the mode).

        The integers will be used to get the corresponding representations from the array.
        The output of this function depends on the current mode.
        If a quantizer is used, this function returns quantized vectors.

        Args:
            ids (Iterable[str]): The document/passage IDs to get the representations for.

        Returns:
            Tuple[np.ndarray, List[List[int]]]: The vectors and corresponding indices.
        """
        pass

    def _compute_scores(
        self, df: pd.DataFrame, query_vectors: np.ndarray
    ) -> pd.DataFrame:
        """Computes scores for a data frame.
        The input data frame needs a "q_no" column with unique query numbers.

        Args:
            df (pd.DataFrame): Input data frame.
            query_vectors (np.ndarray): All query vectors indexed by "q_no".

        Returns:
            pd.DataFrame: Data frame with computed scores.
        """
        # map doc/passage IDs to unique numbers (0 to n)
        id_df = df[["id"]].drop_duplicates().reset_index(drop=True)
        id_df["id_no"] = id_df.index

        # attach doc/passage numbers to data frame
        df = df.merge(id_df, on="id", suffixes=[None, "_"]).reset_index(drop=True)

        # get all required vectors from the FF index
        vectors, id_to_vec_idxs = self._get_vectors(id_df["id"].to_list())
        if self.quantizer is not None:
            vectors = self.quantizer.decode(vectors)

        # compute indices for query vectors and doc/passage vectors in current arrays
        select_query_vectors = []
        select_vectors = []
        select_scores = []
        c = 0
        for id_no, q_no in zip(df["id_no"], df["q_no"]):
            vec_idxs = id_to_vec_idxs[id_no]
            select_vectors.extend(vec_idxs)
            select_scores.append(list(range(c, c + len(vec_idxs))))
            c += len(vec_idxs)
            select_query_vectors.extend([q_no] * len(vec_idxs))

        # compute all dot products (scores)
        q_reps = query_vectors[select_query_vectors]
        d_reps = vectors[select_vectors]
        scores = np.sum(q_reps * d_reps, axis=1)

        # select aggregation operation based on current mode
        if self.mode == Mode.MAXP:
            op = np.max
        elif self.mode == Mode.AVEP:
            op = np.average
        else:
            op = lambda x: x[0]

        def _mapfunc(i):
            scores_i = select_scores[i]
            if len(scores_i) == 0:
                return np.nan
            return op(scores[scores_i])

        # insert FF scores in the correct rows
        df["ff_score"] = df.index.map(_mapfunc)
        return df

    def _early_stopping(
        self,
        df: pd.DataFrame,
        query_vectors: np.ndarray,
        cutoff: int,
        alpha: float,
        intervals: Iterable[int],
    ) -> pd.DataFrame:
        """Compute scores with early stopping for a data frame.
        The input data frame needs a "q_no" column with unique query numbers.

        Args:
            df (pd.DataFrame): Input data frame.
            query_vectors (np.ndarray): All query vectors indexed by "q_no".
            cutoff (int): Cut-off depth for early stopping.
            alpha (float): Interpolation parameter.
            intervals (Iterable[int]): Depths to compute scores at.

        Returns:
            pd.DataFrame: Data frame with computed scores.
        """
        # early stopping splits the data frame, hence we need to keep track of the original index
        df["orig_index"] = df.index

        # data frame for computed scores
        scores_so_far = None

        # [a, b] is the interval for which the scores are computed in each step
        a = 0
        for b in sorted(intervals):
            if b < cutoff:
                continue

            # identify queries which do not meet the early stopping criterion
            if a == 0:
                # first iteration: take all queries
                q_ids_left = pd.unique(df["q_id"])
            else:
                # subsequent iterations: compute ES criterion
                q_ids_left = (
                    scores_so_far.groupby("q_id")
                    .filter(
                        lambda g: g["int_score"].nlargest(cutoff).iat[-1]
                        < alpha * g["score"].iat[-1] + (1 - alpha) * g["ff_score"].max()
                    )["q_id"]
                    .drop_duplicates()
                    .to_list()
                )
            LOGGER.info("depth %s: %s queries left", b, len(q_ids_left))

            # take the next chunk with b-a docs/passages for each query
            chunk = df.loc[df["q_id"].isin(q_ids_left)].groupby("q_id").nth(range(a, b))

            # stop if no pairs are left
            if len(chunk) == 0:
                break

            # compute scores for the chunk and merge
            out = self._compute_scores(chunk, query_vectors)[["orig_index", "ff_score"]]
            chunk_scores = chunk.merge(out, on="orig_index", suffixes=[None, "_"])

            # compute interpolated scores
            chunk_scores["int_score"] = (
                alpha * chunk_scores["score"] + (1 - alpha) * chunk_scores["ff_score"]
            )

            if scores_so_far is None:
                scores_so_far = chunk_scores
            else:
                scores_so_far = pd.concat(
                    [scores_so_far, chunk_scores],
                    axis=0,
                )

            a = b
        return scores_so_far.join(df, on="orig_index", lsuffix=None, rsuffix="_")

    def __call__(
        self,
        ranking: Ranking,
        early_stopping: int = None,
        early_stopping_alpha: float = None,
        early_stopping_intervals: Iterable[int] = None,
    ) -> Ranking:
        """Compute scores for a ranking.

        Args:
            ranking (Ranking): The ranking to compute scores for. Must have queries attached.
            early_stopping (int, optional): Perform early stopping at this cut-off depth. Defaults to None.
            early_stopping_alpha (float, optional): Interpolation parameter for early stopping. Defaults to None.
            early_stopping_intervals (Iterable[int], optional): Intervals for early stopping. Defaults to None.

        Returns:
            Ranking: Ranking with the computed scores.

        Raises:
            ValueError: When the ranking has no queries attached.
            ValueError: When early stopping is enabled but arguments are missing.
        """
        if not ranking.has_queries:
            raise ValueError("Input ranking has no queries attached.")
        if early_stopping is not None and (
            early_stopping_alpha is None or early_stopping_intervals is None
        ):
            raise ValueError("Early stopping requires alpha and intervals.")
        t0 = perf_counter()

        # get all unique queries and query IDs and map to unique numbers (0 to m)
        query_df = (
            ranking._df[["q_id", "query"]].drop_duplicates().reset_index(drop=True)
        )
        query_df["q_no"] = query_df.index

        # attach query numbers to data frame
        df = ranking._df.merge(query_df, on="q_id", suffixes=[None, "_"])

        # batch encode queries
        query_vectors = self.encode_queries(list(query_df["query"]))

        if early_stopping is not None:
            result = self._early_stopping(
                df,
                query_vectors,
                early_stopping,
                early_stopping_alpha,
                early_stopping_intervals,
            )
        else:
            result = self._compute_scores(df, query_vectors)
        result["score"] = result["ff_score"]

        LOGGER.info("computed scores in %s seconds", perf_counter() - t0)
        return Ranking(
            result,
            name="fast-forward",
            dtype=ranking._df.dtypes["score"],
            copy=False,
            is_sorted=False,
        )

    @abc.abstractmethod
    def _batch_iter(
        self, batch_size: int
    ) -> Iterator[Tuple[np.ndarray, IDSequence, IDSequence]]:
        """Iterate over the index in batches (internal method).

        If a quantizer is used, the vectors are the quantized codes.
        When an ID does not exist, it must be set to None.

        Args:
            batch_size (int): Batch size.

        Yields:
            Tuple[np.ndarray, IDSequence, IDSequence]: Vectors, document IDs, passage IDs in batches.
        """
        pass

    def batch_iter(
        self, batch_size: int
    ) -> Iterator[Tuple[np.ndarray, IDSequence, IDSequence]]:
        """Iterate over all vectors, document IDs, and passage IDs in batches.
        IDs may be either strings or None.

        Args:
            batch_size (int): Batch size.

        Yields:
            Tuple[np.ndarray, IDSequence, IDSequence]: Batches of vectors, document IDs (if any), passage IDs (if any).
        """
        if self._quantizer is None:
            yield from self._batch_iter(batch_size)

        else:
            for batch in self._batch_iter(batch_size):
                yield self._quantizer.decode(batch[0]), batch[1], batch[2]

    def __iter__(
        self,
    ) -> Iterator[Tuple[np.ndarray, Optional[str], Optional[str]]]:
        """Iterate over all vectors, document IDs, and passage IDs.

        Yields:
            Tuple[np.ndarray, Optional[str], Optional[str]]: Vector, document ID (if any), passage ID (if any).
        """
        for vectors, doc_ids, psg_ids in self.batch_iter(2**9):
            yield from zip(vectors, doc_ids, psg_ids)



"""
.. include:: docs/ranking.md
"""

import logging
from pathlib import Path
from typing import Dict, Iterator, Set, Union

import numpy as np
import pandas as pd

Run = Dict[str, Dict[str, Union[float, int]]]

LOGGER = logging.getLogger(__name__)


def _attach_queries(df: pd.DataFrame, queries: Dict[str, str]) -> pd.DataFrame:
    """Attach queries to a data frame.

    Args:
        df (pd.DataFrame): The data frame (same format as used by rankings).
        queries (Dict[str, str]): Query IDs mapped to queries.

    Raises:
        ValueError: When the queries are incomplete.

    Returns:
        pd.DataFrame: The data frame with queries attached.
    """
    if not set(pd.unique(df["q_id"])).issubset(set(queries.keys())):
        raise ValueError("Queries are incomplete.")
    return df.merge(
        pd.DataFrame(queries.items(), columns=["q_id", "query"]), how="left", on="q_id"
    )


class Ranking(object):
    """Represents rankings of documents/passages w.r.t. queries."""

    def __init__(
        self,
        df: pd.DataFrame,
        name: str = None,
        queries: Dict[str, str] = None,
        dtype: np.dtype = np.float32,
        copy: bool = True,
        is_sorted: bool = False,
    ) -> None:
        """Create a ranking from an existing data frame. Removes rows with NaN scores.

        Args:
            df (pd.DataFrame): Data frame containing IDs and scores.
            name (str, optional): Method name. Defaults to None. Defaults to None.
            queries (Dict[str, str], optional): Query IDs mapped to queries. Defaults to None.
            dtype (np.dtype, optional): How the scores should be represented in the data frame. Defaults to np.float32.
            copy (bool, optional): Whether to copy the data frame. Defaults to True.
            is_sorted (bool, optional): Whether the data frame is already sorted (by score). Defaults to False.

        Raises:
            ValueError: When the queries are incomplete.
        """
        super().__init__()
        self.name = name

        cols = ["q_id", "id", "score"]
        if "query" in df.columns:
            cols += ["query"]
        self._df = df.loc[:, cols].dropna()
        if copy:
            self._df = self._df.copy()

        for col, dt in (
            ("score", dtype),
            ("q_id", str),
            ("id", str),
        ):
            if col in self._df.columns and self._df[col].dtype != dt:
                self._df[col] = self._df[col].astype(dt)

        if not is_sorted:
            self._df.sort_values(by=["q_id", "score"], inplace=True, ascending=False)
        self._df.reset_index(inplace=True, drop=True)

        self._q_ids = set(pd.unique(self._df["q_id"]))
        if queries is not None:
            self._df = _attach_queries(self._df, queries)

    @property
    def has_queries(self) -> bool:
        """Whether the ranking has queries attached.

        Returns:
            bool: Whether queries exist.
        """
        return "query" in self._df.columns

    @property
    def q_ids(self) -> Set[str]:
        """The set of (unique) query IDs in this ranking. Only queries with at least one scored document are considered.

        Returns:
            Set[str]: The query IDs.
        """
        return self._q_ids

    def __getitem__(self, q_id: str) -> Dict[str, float]:
        """Return the ranking for a query.

        Args:
            q_id (str): The query ID.

        Returns:
            Dict[str, float]: Document/passage IDs mapped to scores.
        """
        return dict(self._df[self._df["q_id"] == q_id][["id", "score"]].values)

    def __len__(self) -> int:
        """Return the number of queries.

        Returns:
            int: The number of queries.
        """
        return len(self._q_ids)

    def __iter__(self) -> Iterator[str]:
        """Yield all query IDs.

        Yields:
            str: The query IDs.
        """
        yield from self._q_ids

    def __contains__(self, key: object) -> bool:
        """Check whether a query ID is in the ranking.

        Args:
            key (object): The query ID.

        Returns:
            bool: Whether the query ID has associated document/passage IDs.
        """
        return key in self._q_ids

    def __eq__(self, o: object) -> bool:
        """Check if this ranking is identical to another one. Only takes IDs and scores into account.

        Args:
            o (object): The other ranking.

        Returns:
            bool: Whether the two rankings are identical.
        """
        if not isinstance(o, Ranking):
            return False

        df1 = self._df.sort_values(["q_id", "id"]).reset_index(drop=True)
        df2 = o._df.sort_values(["q_id", "id"]).reset_index(drop=True)

        cols = ["q_id", "id", "score"]
        return df1[cols].equals(df2[cols])

    def __repr__(self) -> str:
        """Return a string representation of this ranking.

        Returns:
            str: The string representation.
        """
        return self._df.__repr__()

    def attach_queries(self, queries: Dict[str, str]) -> "Ranking":
        """Attach queries to the ranking.

        Args:
            queries (Dict[str, str]): Query IDs mapped to queries.

        Raises:
            ValueError: When the queries are incomplete.

        Returns:
            Ranking: The ranking with queries attached.
        """
        return Ranking(
            self._df,
            self.name,
            queries=queries,
            dtype=self._df.dtypes["score"],
            copy=True,
            is_sorted=True,
        )

    def cut(self, cutoff: int) -> "Ranking":
        """For each query, remove all but the top-`k` scoring documents/passages.

        Args:
            cutoff (int): Number of best scores per query to keep (`k`).

        Returns:
            Ranking: The resulting ranking.
        """
        return Ranking(
            self._df.groupby("q_id").head(cutoff).reset_index(drop=True),
            name=self.name,
            dtype=self._df.dtypes["score"],
            copy=True,
            is_sorted=True,
        )

    def interpolate(
        self,
        other: "Ranking",
        alpha: float,
    ) -> "Ranking":
        """Interpolate as `score = self.score * alpha + other.score * (1 - alpha)`.

        Args:
            other (Ranking): Ranking to interpolate with.
            alpha (float): Interpolation parameter.

        Returns:
            Ranking: The resulting ranking.
        """
        # preserves order by score
        new_df = self._df.merge(other._df, on=["q_id", "id"], suffixes=[None, "_other"])
        new_df["score"] = alpha * new_df["score"] + (1 - alpha) * new_df["score_other"]
        result = Ranking(
            new_df,
            name=self.name,
            dtype=self._df.dtypes["score"],
            copy=False,
            is_sorted=False,
        )
        return result

    def save(
        self,
        target: Path,
    ) -> None:
        """Save the ranking in a TREC runfile.

        Args:
            target (Path): Output file.
        """
        df_ranks = self._df.groupby("q_id").cumcount().to_frame()
        df_ranks.columns = ("rank",)
        df_out = self._df.join(df_ranks)
        df_out["name"] = str(self.name)
        df_out["q0"] = "Q0"

        target.parent.mkdir(parents=True, exist_ok=True)
        df_out.to_csv(
            target,
            sep="\t",
            columns=["q_id", "q0", "id", "rank", "score", "name"],
            index=False,
            header=False,
        )

    @classmethod
    def from_run(
        cls,
        run: Run,
        name: str = None,
        queries: Dict[str, str] = None,
        dtype: np.dtype = np.float32,
    ) -> "Ranking":
        """Create a Ranking object from a TREC run.

        Args:
            run (Run): TREC run.
            name (str, optional): Method name. Defaults to None. Defaults to None.
            queries (Dict[str, str], optional): Query IDs mapped to queries. Defaults to None.
            dtype (np.dtype, optional): How the score should be represented in the data frame. Defaults to np.float32.

        Returns:
            Ranking: The resulting ranking.
        """
        df = pd.DataFrame.from_dict(run).stack().reset_index()
        df.columns = ("id", "q_id", "score")
        return cls(df, name=name, queries=queries, dtype=dtype, copy=False)

    @classmethod
    def from_file(
        cls,
        f: Path,
        queries: Dict[str, str] = None,
        dtype: np.dtype = np.float32,
    ) -> "Ranking":
        """Create a Ranking object from a runfile in TREC format.

        Args:
            f (Path): TREC runfile to read.
            queries (Dict[str, str], optional): Query IDs mapped to queries. Defaults to None.
            dtype (np.dtype, optional): How the score should be represented in the data frame. Defaults to np.float32.

        Returns:
            Ranking: The resulting ranking.
        """
        df = pd.read_csv(
            f,
            sep=r"\s+",
            skipinitialspace=True,
            header=None,
            names=["q_id", "q0", "id", "rank", "score", "name"],
        )
        return cls(df, name=df["name"][0], queries=queries, dtype=dtype, copy=False)



import unittest

import numpy as np

from fast_forward.quantizer import Quantizer
from fast_forward.quantizer.nanopq import NanoOPQ, NanoPQ


class TestQuantizer(unittest.TestCase):
    __test__ = False

    def test_eq(self):
        self.assertEqual(self.quantizer, self.quantizer)
        self.assertEqual(self.quantizer_trained, self.quantizer_trained)
        self.assertNotEqual(self.quantizer, self.quantizer_trained)

    def test_properties(self):
        self.assertEqual((None, 8), self.quantizer.dims)
        self.assertEqual(np.uint8, self.quantizer.dtype)
        self.assertFalse(self.quantizer._trained)

        self.assertEqual((768, 8), self.quantizer_trained.dims)
        self.assertEqual(np.uint8, self.quantizer_trained.dtype)
        self.assertTrue(self.quantizer_trained._trained)

    def test_encoding_decoding(self):
        inputs = np.random.normal(size=(8, 768)).astype(np.float32)
        encoded = self.quantizer_trained.encode(inputs)
        self.assertEqual(encoded.shape, (8, 8))
        self.assertEqual(encoded.dtype, np.uint8)
        decoded = self.quantizer_trained.decode(encoded)
        self.assertEqual(decoded.shape, inputs.shape)

    def test_serialization(self):
        inputs = np.random.normal(size=(8, 768)).astype(np.float32)
        quantizer_loaded = Quantizer.deserialize(*self.quantizer.serialize())
        self.assertEqual(self.quantizer, quantizer_loaded)

        quantizer_trained_loaded = Quantizer.deserialize(
            *self.quantizer_trained.serialize()
        )
        self.assertEqual(self.quantizer_trained, quantizer_trained_loaded)
        np.testing.assert_array_equal(
            self.quantizer_trained.encode(inputs),
            quantizer_trained_loaded.encode(inputs),
        )

    def test_errors(self):
        # encoding before the quantizer is trained
        with self.assertRaises(RuntimeError):
            self.quantizer.encode(np.random.normal(size=(8, 768)).astype(np.float32))

        # attaching to index before the quantizer is trained
        with self.assertRaises(RuntimeError):
            self.quantizer.set_attached()


class TestNanoPQ(TestQuantizer):
    __test__ = True

    @classmethod
    def setUpClass(self):
        self.quantizer = NanoPQ(8, 256)
        self.quantizer_trained = NanoPQ(8, 256)
        self.quantizer_trained.fit(
            np.random.normal(size=(2**10, 768)).astype(np.float32)
        )


class TestNanoOPQ(TestQuantizer):
    __test__ = True

    @classmethod
    def setUpClass(self):
        self.quantizer = NanoOPQ(8, 256)
        self.quantizer_trained = NanoOPQ(8, 256)
        self.quantizer_trained.fit(
            np.random.normal(size=(2**10, 768)).astype(np.float32)
        )


if __name__ == "__main__":
    unittest.main()



"""
.. include:: ../docs/quantizer.md
"""

import abc
import importlib
import logging
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np

LOGGER = logging.getLogger(__name__)


QuantizerAttributes = Dict[str, Union[str, bool, int, float]]
QuantizerData = Dict[str, np.ndarray]


class Quantizer(abc.ABC):
    """Base class for quantizers."""

    _attached: bool = False
    _trained: bool = False

    def __eq__(self, o: object) -> bool:
        """Check whether this quantizer is identical to another one.

        Args:
            o (object): The other quantizer.

        Returns:
            bool: Whether the two quantizers are identical.
        """
        if not isinstance(o, Quantizer):
            return False

        self_meta, self_attributes, self_data = self.serialize()
        o_meta, o_attributes, o_data = o.serialize()

        if self_meta != o_meta or self_attributes != o_attributes:
            return False

        if self_data.keys() != o_data.keys():
            return False

        for k, v in self_data.items():
            if (v != o_data[k]).any():
                return False

        return True

    def set_attached(self) -> None:
        """Set the quantizer as attached, preventing calls to `Quantizer.fit`."""
        if not self._trained:
            raise RuntimeError(
                f"Call {self.__class__.__name__}.fit before attaching the quantizer to an index."
            )
        self._attached = True

    @abc.abstractmethod
    def _fit(self, vectors: np.ndarray, **kwargs: Any) -> None:
        """Fit the quantizer (internal method).

        Args:
            vectors (np.ndarray): The training vectors.
            **kwargs (Any): Arguments specific to the quantizer.
        """
        pass

    def fit(self, vectors: np.ndarray, **kwargs: Any) -> None:
        """Fit (train) the quantizer.

        Quantizers can only be trained before being attached to an index to avoid inconsistencies.

        Args:
            vectors (np.ndarray): The training vectors.
            **kwargs (Any): Arguments specific to the quantizer.

        Raises:
            RuntimeError: When the quantizer is aready attached to an index.
        """
        if self._attached:
            raise RuntimeError(
                "Quantizers can only be fitted before they are attached to an index."
            )
        self._fit(vectors, **kwargs)
        self._trained = True

    @property
    @abc.abstractmethod
    def dtype(self) -> np.dtype:
        """The data type of the codes produced by this quantizer.

        Returns:
            np.dtype: The data type.
        """
        pass

    @property
    @abc.abstractmethod
    def dims(self) -> Tuple[Optional[int], Optional[int]]:
        """The dimensions before and after quantization.

        May return None values before the quantizer is trained.

        Returns:
            Tuple[Optional[int], Optional[int]]: Dimension of the original vectors and dimension of the codes.
        """
        pass

    @abc.abstractmethod
    def _encode(self, vectors: np.ndarray) -> np.ndarray:
        """Encode vectors (internal method).

        Args:
            vectors (np.ndarray): The vectors to be encoded.

        Returns:
            np.ndarray: The codes corresponding to the vectors.
        """
        pass

    def encode(self, vectors: np.ndarray) -> np.ndarray:
        """Encode a batch of vectors.

        Args:
            vectors (np.ndarray): The vectors to be encoded.

        Returns:
            np.ndarray: The codes corresponding to the vectors.
        """
        if not self._trained:
            raise RuntimeError(f"Call {self.__class__.__name__}.fit first.")
        return self._encode(vectors)

    @abc.abstractmethod
    def _decode(self, codes: np.ndarray) -> np.ndarray:
        """Reconstruct vectors (internal method).

        Args:
            codes (np.ndarray): The codes to be decoded.

        Returns:
            np.ndarray: The reconstructed vectors.
        """
        pass

    def decode(self, codes: np.ndarray) -> np.ndarray:
        """Decode a batch of codes to obtain approximate vector representations.

        Args:
            codes (np.ndarray): The codes to be decoded.

        Returns:
            np.ndarray: The approximated vectors.
        """
        if not self._trained:
            raise RuntimeError(f"Call {self.__class__.__name__}.fit first.")
        return self._decode(codes)

    @abc.abstractmethod
    def _get_state(self) -> Tuple[QuantizerAttributes, QuantizerData]:
        """Return key-value pairs that represent the state of the quantizer (internal method).

        This method returns a tuple of quantizer attributes (values) and quantizer data (numpy arrays).

        Returns:
            Tuple[QuantizerAttributes, QuantizerData]: Attributes and data of the quantizer.
        """
        pass

    def serialize(
        self,
    ) -> Tuple[QuantizerAttributes, QuantizerAttributes, QuantizerData]:
        """Return a serialized representation of the quantizer that can be stored in the index.

        Returns:
            Tuple[QuantizerAttributes, QuantizerAttributes, QuantizerData]: The serialized quantizer.
        """
        meta = {
            "__module__": self.__class__.__module__,
            "__name__": self.__class__.__name__,
            "_trained": self._trained,
        }
        attributes, data = self._get_state()
        return meta, attributes, data

    @classmethod
    @abc.abstractmethod
    def _from_state(
        cls, attributes: QuantizerAttributes, data: QuantizerData
    ) -> "Quantizer":
        """Instantiate a quantizer based on its state.

        Args:
            attributes (QuantizerAttributes): The quantizer attributes.
            data (QuantizerData): The quantizer data.

        Returns:
            Quantizer: The resulting quantizer.
        """
        pass

    @classmethod
    def deserialize(
        cls,
        meta: QuantizerAttributes,
        attributes: QuantizerAttributes,
        data: QuantizerData,
    ) -> "Quantizer":
        """Reconstruct a serialized quantizer.

        Args:
            meta (QuantizerAttributes): The quantizer metadata.
            attributes (QuantizerAttributes): The quantizer attributes.
            data (QuantizerData): The quantizer data.

        Returns:
            Quantizer: The loaded quantizer.
        """
        LOGGER.debug("reconstructing %s.%s", meta["__module__"], meta["__name__"])
        quantizer_mod = importlib.import_module(meta["__module__"])
        quantizer_cls = getattr(quantizer_mod, meta["__name__"])
        quantizer = quantizer_cls._from_state(attributes, data)
        quantizer._trained = meta["_trained"]
        return quantizer



import pandas as pd
import pyterrier as pt

from fast_forward.index import Index
from fast_forward.ranking import Ranking


class FFScore(pt.Transformer):
    """PyTerrier transformer that computes scores using a Fast-Forward index."""

    def __init__(self, index: Index) -> None:
        """Create an FFScore transformer.

        Args:
            index (Index): The Fast-Forward index.
        """
        self._index = index
        super().__init__()

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """Compute the scores for all query-document pairs in the data frame.
        The previous scores are moved to the "score_0" column.

        Args:
            df (pd.DataFrame): The PyTerrier data frame.

        Returns:
            pd.DataFrame: A new data frame with the computed scores.
        """
        ff_scores = self._index(
            Ranking(
                df.rename(columns={"qid": "q_id", "docno": "id"}),
                copy=False,
                is_sorted=True,
            )
        )._df.rename(columns={"q_id": "qid", "id": "docno"})

        return df[["qid", "docno", "score"]].merge(
            ff_scores[["qid", "docno", "score", "query"]],
            on=["qid", "docno"],
            suffixes=["_0", None],
        )

    def __repr__(self) -> str:
        """Return a string representation.
        The representation is unique w.r.t. the index and its query encoder.

        Returns:
            str: The representation.
        """
        return f"{self.__class__.__name__}({id(self._index)}, {id(self._index._query_encoder)})"


class FFInterpolate(pt.Transformer):
    """PyTerrier transformer that interpolates scores computed by `FFScore`."""

    def __init__(self, alpha: float) -> None:
        """Create an FFInterpolate transformer.

        Args:
            alpha (float): The interpolation parameter.
        """
        # attribute name needs to be exactly this for pyterrier.GridScan to work
        self.alpha = alpha
        super().__init__()

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """Interpolate the scores for all query-document pairs in the data frame as
        `alpha * score_0 + (1 - alpha) * score`.

        Args:
            df (pd.DataFrame): The PyTerrier data frame.

        Returns:
            pd.DataFrame: A new data frame with the interpolated scores.
        """
        new_df = df[["qid", "docno", "query"]].copy()
        new_df["score"] = self.alpha * df["score_0"] + (1 - self.alpha) * df["score"]
        return new_df



"""
.. include:: docs/main.md
"""

import importlib.metadata

from fast_forward.index import Mode
from fast_forward.index.disk import OnDiskIndex
from fast_forward.index.memory import InMemoryIndex
from fast_forward.indexer import Indexer
from fast_forward.ranking import Ranking

__version__ = importlib.metadata.version("fast-forward-indexes")



import unittest

import numpy as np

from fast_forward.encoder import (
    LambdaEncoder,
    TCTColBERTDocumentEncoder,
    TCTColBERTQueryEncoder,
)


class TestTCTColBERTEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.query_encoder = TCTColBERTQueryEncoder(
            "castorini/tct_colbert-msmarco", device="cpu"
        )
        self.doc_encoder = TCTColBERTDocumentEncoder(
            "castorini/tct_colbert-msmarco", device="cpu"
        )

    def test_query_encoder(self):
        out = self.query_encoder(["test query 1", "test query 2"])
        self.assertEqual(out.shape, (2, 768))

    def test_doc_encoder(self):
        out = self.doc_encoder(["test doc 1", "test doc 11"])
        self.assertEqual(out.shape, (2, 768))


class TestLambdaEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.encoder = LambdaEncoder(lambda q: np.zeros(shape=(768,)))

    def test_encode(self):
        q_enc = self.encoder(["test query 1", "test query 2"])
        self.assertTrue(np.array_equal(q_enc, np.zeros(shape=(2, 768))))


if __name__ == "__main__":
    unittest.main()



import unittest

from fast_forward import Ranking
from fast_forward.util import to_ir_measures

from .test_ranking import DUMMY_QUERIES, RUN


class TestUtil(unittest.TestCase):
    def test_ir_measures_df(self):
        r = Ranking.from_run(RUN, queries=DUMMY_QUERIES)
        df = to_ir_measures(r)
        self.assertTrue(df["query_id"].equals(r._df["q_id"]))
        self.assertTrue(df["doc_id"].equals(r._df["id"]))
        self.assertTrue(df["score"].equals(r._df["score"]))
        self.assertEqual(set(df.columns), set(("query_id", "doc_id", "score")))


if __name__ == "__main__":
    unittest.main()



"""
.. include:: docs/indexer.md
"""

from typing import Dict, Iterable

from tqdm import tqdm

from fast_forward.encoder import Encoder
from fast_forward.index import Index


class Indexer(object):
    """Utility class for indexing collections."""

    def __init__(self, index: Index, encoder: Encoder, batch_size: int = 32) -> None:
        """Constructor.

        Args:
            index (Index): The index to add the collection to.
            encoder (Encoder): Document/passage encoder.
            batch_size (int, optional): Batch size for encoding. Defaults to 32.
        """
        self._index = index
        self._encoder = encoder
        self._batch_size = batch_size

    def index_dicts(self, data: Iterable[Dict[str, str]]) -> None:
        """Index data from dictionaries.

        The dictionaries should have the key "text" and at least one of "doc_id" and "psg_id".

        Args:
            data (Iterable[Dict[str, str]]): An iterable of the dictionaries.
        """
        texts, doc_ids, psg_ids = [], [], []
        for d in tqdm(data):
            texts.append(d["text"])
            doc_ids.append(d.get("doc_id"))
            psg_ids.append(d.get("psg_id"))

            if len(texts) == self._batch_size:
                self._index.add(self._encoder(texts), doc_ids=doc_ids, psg_ids=psg_ids)
                texts, doc_ids, psg_ids = [], [], []

        if len(texts) > 0:
            self._index.add(self._encoder(texts), doc_ids=doc_ids, psg_ids=psg_ids)



import unittest

import numpy as np

from fast_forward.encoder import LambdaEncoder
from fast_forward.index.memory import InMemoryIndex
from fast_forward.indexer import Indexer


class TestIndexer(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.index = InMemoryIndex()
        self.indexer = Indexer(
            self.index, LambdaEncoder(lambda q: np.zeros(shape=(16,))), batch_size=2
        )

    def test_index_dicts(self):
        self.indexer.index_dicts(
            [
                {"text": "123", "doc_id": "d1", "psg_id": "d1_p1"},
                {"text": "234", "doc_id": "d1", "psg_id": "d1_p2"},
                {"text": "456", "doc_id": "d1", "psg_id": "d1_p3"},
                {"text": "567", "doc_id": "d2", "psg_id": "d2_p1"},
                {"text": "678", "doc_id": "d3", "psg_id": "d3_p1"},
                {"text": "890", "doc_id": "d4"},
                {"text": "901", "psg_id": "d5_p1"},
            ]
        )
        self.assertEqual(7, len(self.index))
        self.assertEqual(set(("d1", "d2", "d3", "d4")), self.index.doc_ids)
        self.assertEqual(
            set(("d1_p1", "d1_p2", "d1_p3", "d2_p1", "d3_p1", "d5_p1")),
            self.index.psg_ids,
        )


if __name__ == "__main__":
    unittest.main()
